#ifndef __toupnam_h__
#define __toupnam_h__

/* Version: 5.13010.20180926 */
/*
   Platform & Architecture:
            (1) Win32:
                  (a) x86: XP SP3 or above; CPU supports SSE2 instruction set or above
                  (b) x64: Win7 or above
            (2) macOS: x64; macOS 10.10 or above
            (3) Linux: x86, x64; kernel 2.6.27 or above; GLIBC 2.17 or above
*/

#ifdef _WIN32
#ifndef _INC_WINDOWS
#include <windows.h>
#endif
#endif

#ifdef __cplusplus
extern "C" {
#endif

#ifdef _WIN32
#pragma pack(push, 8)
#ifdef TOUPNAM_EXPORTS
#define toupnam_ports(x)    __declspec(dllexport)   x   __stdcall  /* in Windows, we use __stdcall calling convention, see https://docs.microsoft.com/en-us/cpp/cpp/stdcall */
#elif !defined(TOUPNAM_NOIMPORTS)
#define toupnam_ports(x)    __declspec(dllimport)   x   __stdcall
#else
#define toupnam_ports(x)    x   __stdcall
#endif

#else

#define toupnam_ports(x)    x

#ifndef HRESULT
#define HRESULT int
#endif

#ifndef __stdcall
#define __stdcall
#endif

#ifndef CALLBACK
#define CALLBACK
#endif

#ifndef BOOL
#define BOOL int
#endif

#ifndef __BITMAPINFOHEADER_DEFINED__
#define __BITMAPINFOHEADER_DEFINED__
typedef struct {
    unsigned        biSize;
    int             biWidth;
    int             biHeight;
    unsigned short  biPlanes;
    unsigned short  biBitCount;
    unsigned        biCompression;
    unsigned        biSizeImage;
    int             biXPelsPerMeter;
    int             biYPelsPerMeter;
    unsigned        biClrUsed;
    unsigned        biClrImportant;
} BITMAPINFOHEADER;
#endif

#endif

/* handle */
typedef struct ToupnamT { int unused; } *HToupNam;

#ifndef __TOUPNAMINST_DEFINED__
#define __TOUPNAMINST_DEFINED__

#define TOUPNAM_MAX         32

typedef struct{
    int idisable;   /* 0 = "support this feature", 1 = "not support" */
    int imin; /* minimum value */
    int imax; /* maximum value */
    int idef; /* default value */
}ToupnamRange;

#define TOUPNAM_PARA_UNKNOWN                0x00
#define TOUPNAM_PARA_EXPOTIME               0x01    /* exposure time */
#define TOUPNAM_PARA_AGAIN                  0x02    /* analog gain */
#define TOUPNAM_PARA_AEXPOTARGET            0x03    /* auto exposure target */
#define TOUPNAM_PARA_TEMP                   0x04    /* color temperature */
#define TOUPNAM_PARA_TINT                   0x05
#define TOUPNAM_PARA_CONTRAST               0x06    /* contrast */
#define TOUPNAM_PARA_HUE                    0x07    /* hue */
#define TOUPNAM_PARA_SATURATION             0x08    /* saturation */
#define TOUPNAM_PARA_BRIGHTNESS             0x09    /* brightness */
#define TOUPNAM_PARA_GAMMA                  0x0a    /* gamma */
#define TOUPNAM_PARA_AEXPO                  0x0b    /* auto exposure */
#define TOUPNAM_PARA_AWB                    0x0c    /* auto white balance */
#define TOUPNAM_PARA_BINSKIP                0x0d    /* bin / skip */
#define TOUPNAM_PARA_HZ                     0x0e    /* power supply: 0 -> 60HZ AC;  1 -> 50Hz AC;   2 -> DC */
#define TOUPNAM_PARA_BPS                    0x0f    /* bits per second, kbps */
#define TOUPNAM_PARA_KEYFRAME               0x10    /* key frame interval */
#define TOUPNAM_PARA_LOWLIGHTCOMPENSATION   0x11    /* low light compensation */
#define TOUPNAM_PARA_SHARPNESS              0x12    /* sharpness */
#define TOUPNAM_PARA_WBREDGAIN              0x13    /* white balance red gain */
#define TOUPNAM_PARA_WBGREENGAIN            0x14    /* white balance green gain */
#define TOUPNAM_PARA_WBBLUEGAIN             0x15    /* white balance blue gain */
#define TOUPNAM_PARA_DENOISE                0x16    /* denoise */
#define TOUPNAM_PARA_APSTA                  0x17    /* ap/sta */
#define TOUPNAM_PARA_CODEC                  0x18    /* codec, H264, H265, etc */
#define TOUPNAM_PARA_AFPOSITION             0x19    /* af sensor board positon */
#define TOUPNAM_PARA_AFMODE                 0x1a    /* af mode (0:manul focus; 1:auto focus; 2:onepush focus; 3:conjugate calibration) */
#define TOUPNAM_PARA_AFZONE                 0x1b    /* af zone */
#define TOUPNAM_PARA_AFFEEDBACK             0x1c    /* af information feedback; 0:unknown; 1:focused; 2:focusing; 3:defocus; 4:up; 5:down */
#define TOUPNAM_PARA_AFPOSITION_ABSOLUTE    0x1d    /* absolute af sensor board positon */
#define TOUPNAM_PARA_STATUS                 0x1e    /* status */
#define TOUPNAM_PARA_EVENT                  0x1f    /* event */

#define TOUPNAM_PARA_CHROME                 0x80    /* monochromatic mode */
#define TOUPNAM_PARA_VFLIP                  0x81    /* vertical flip */
#define TOUPNAM_PARA_HFLIP                  0x82    /* horizontal flip */
#define TOUPNAM_PARA_NEGATIVE               0x83    /* negative film */

#define TOUPNAM_PARA_RETRY                  0xe0

#define TOUPNAM_PARA_STATUS_RECORDING       0x00000001      /* recording */ 
#define TOUPNAM_PARA_STATUS_SD              0x00000002      /* sd card available */
#define TOUPNAM_PARA_STATUS_SD_FULL         0x00000004      /* sd card full */

#define TOUPNAM_PARA_EVENT_FAT4G            0x00000001      /* file size limit 4g in FAT32 */

#define TOUPNAM_STATE_INITING               0x00    /* initialization */
#define TOUPNAM_STATE_NORMAL                0x01    /* normal */
#define TOUPNAM_STATE_UNREACHABLE           0x02    /* network not reachable */

#define TOUPNAM_FLAG_WIFI_AP        0x0001
#define TOUPNAM_FLAG_WIFI_STA       0x0002
#define TOUPNAM_FLAG_ETHERNET       0x0004
#define TOUPNAM_FLAG_CAPTURE        0x0008  /* support the ability of capture image from camera */
#define TOUPNAM_FLAG_AWBCHECKMODE   0x0010  /* auto white balance: check mode vs 'one push' mode */
#define TOUPNAM_FLAG_UVC            0x0020  /* uvc camera */
#define TOUPNAM_FLAG_WBGAIN         0x0040  /* white balance gain mode or temp tint mode */
#define TOUPNAM_FLAG_MULTICAST      0x0080  /* RTSP/RTP multicast */
#define TOUPNAM_FLAG_AF             0x0100  /* support auto focus */
#define TOUPNAM_FLAG_SD_LIST        0x0200  /* support to list sd card */

typedef struct{
    char            id[64];         /* unique id, used for Toupnam_Open */
    char            sn[64];         /* serial number */
    char            name[64];
    char            model[64];
    char            version[64];
    char            addr[64];       /* ip */
    char            url[256];       /* playback url, such as rtsp://xxxx/yyyy */
    unsigned        state;  /* TOUPNAM_STATE_xxx */
    unsigned        flag;   /* TOUPNAM_FLAG_xxx */
    ToupnamRange    range[TOUPNAM_MAX];
}ToupnamInst;

#endif

typedef struct {
    char    ssid[64];
    char    password[64];
}ToupnamWifi;

/*
    get the version of this dll/so/dylib, which is: 5.13010.20180926
*/
toupnam_ports(const char*)      Toupnam_Version();

#define TOUPNAM_EVENT_ENUM        0x01    /* enum */
#define TOUPNAM_EVENT_WIFI        0x02    /* wifi */
#define TOUPNAM_EVENT_PARA        0x03    /* parameter change TOUPNAM_PARA_xxxx */
#define TOUPNAM_EVENT_IMAGE       0x04    /* image */
#define TOUPNAM_EVENT_LISTWIFI    0x05    /* list wifi finished */
#define TOUPNAM_EVENT_LISTDIR     0x07    /* list dir */
#define TOUPNAM_EVENT_THUMBNAIL   0x08    /* thumbnail */
#define TOUPNAM_EVENT_DIRCHANGE   0x09    /* dir change notify */
#define TOUPNAM_EVENT_RECORDSTART 0x0a    /* record start */
#define TOUPNAM_EVENT_RECORDSTOP  0x0b    /* record stop */
#define TOUPNAM_EVENT_DATETIME    0x0c    /* date time */
#define TOUPNAM_EVENT_ERROR       0x80    /* error */

typedef struct{
    int         result;
    void*       ctx;
    void*       ptr;
    unsigned    length;
}ToupnamEventExtra;

typedef void (__stdcall* PTOUPNAM_EVENT_CALLBACK)(unsigned nEvent, unsigned nPara, void* pCallbackCtx, ToupnamEventExtra* pExtra);
/*
    when toupnam.dll/libtoupnam.so discovery new camera, pCallback will be called.
    pCallback can be NULL if the application does not interest this.
    Toupnam_Init: call only once when application startup
    Toupnam_Fini: call only once when application exit
*/
toupnam_ports(void)     Toupnam_Init(PTOUPNAM_EVENT_CALLBACK pCallback, void* pCallbackCtx);
toupnam_ports(void)     Toupnam_Fini();

toupnam_ports(unsigned) Toupnam_Enum(ToupnamInst pti[TOUPNAM_MAX]); /* enumerate the cameras discovered by the computer, return the number of enumerated */

toupnam_ports(HToupNam) Toupnam_Open(const char* id);
toupnam_ports(HToupNam) Toupnam_Open_ByIndex(unsigned index);
toupnam_ports(void)     Toupnam_Close(HToupNam h); /* close the handle */

/*
    when a frame is arrived, pCallbackCtx is callbacked. (NULL == pData) means that something is error.
    when TOUPNAM_PARA_xxx value is changed, pParaCallback is callbacked.
    pCallbackCtx is the callback context which is passed by Toupcam_Start.
    pDataCallback and pParaCallback are callbacked by an internal thread of toupnam.dll, so please pay attention to multithread problem.
*/
typedef void (__stdcall* PTOUPNAM_PARA_CALLBACK)(unsigned para, int value, void* pCallbackCtx);
typedef void (__stdcall* PTOUPNAM_DATA_CALLBACK)(const void* pData, const BITMAPINFOHEADER* pHeader, void* pCallbackCtx);
typedef void (__stdcall* PTOUPNAM_CAPTURE_CALLBACK)(int result, const void* pData, size_t nLength, const BITMAPINFOHEADER* pHeader, void* pCallbackCtx);

toupnam_ports(HRESULT)  Toupnam_StartPushMode(HToupNam h, PTOUPNAM_DATA_CALLBACK pDataCallback, PTOUPNAM_PARA_CALLBACK pParaCallback, void* pCallbackCtx);

#ifdef _WIN32
toupnam_ports(HRESULT)  Toupnam_StartPullModeWithWndMsg(HToupNam h, HWND hWnd, UINT nMsg);
#endif

/*
    bits: 24 (RGB24), 32 (RGB32), or 8 (Grey)
*/
toupnam_ports(HRESULT)  Toupnam_PullImage(HToupNam h, void* pImageData, int bits, unsigned* pnWidth, unsigned* pnHeight);
toupnam_ports(HRESULT)  Toupnam_StartPullModeWithCallback(HToupNam h, void* pCallbackContext);

toupnam_ports(HRESULT)  Toupnam_Stop(HToupNam h);
toupnam_ports(HRESULT)  Toupnam_Pause(HToupNam h, BOOL bPause);

/* capture image, compare this to image extracted from video
   pstrFileName:
                NULL        -> capture image and then return by callback
                "abc.jpg"   -> capture image and then save it in the camera sd card with filename 'abc.jpg'
                "thumbnail" -> capture the thumbnail image and then return by callback
                "*"         -> capture image and then save it in the camera sd card with auto generated file name
*/
toupnam_ports(HRESULT)  Toupnam_Capture(HToupNam h, const char* pstrFileName, PTOUPNAM_CAPTURE_CALLBACK pCaptureCallback, void* pCallbackCtx);
toupnam_ports(HRESULT)  Toupnam_Capture_ById(const char* id, const char* pstrFileName, PTOUPNAM_CAPTURE_CALLBACK pCaptureCallback, void* pCallbackCtx);

toupnam_ports(ToupnamInst)  Toupnam_get_Inst(HToupNam h);

toupnam_ports(HRESULT)  Toupnam_get_Size(HToupNam h, int* pWidth, int* pHeight);
toupnam_ports(HRESULT)  Toupnam_get_CapSize(HToupNam h, int* pWidth, int* pHeight);
toupnam_ports(HRESULT)  Toupnam_get_FourCC(HToupNam h, unsigned* pFourCC); /* see http://en.wikipedia.org/wiki/FourCC, http://www.fourcc.org */

/*
    (fullpathname == NULL) means to stop record.
    support file extension: *.asf, *.mp4, *.mkv
*/
toupnam_ports(HRESULT)  Toupnam_Record(HToupNam h, const char* fullpathname);

toupnam_ports(HRESULT)  Toupnam_put_Para(HToupNam h, unsigned para, int value); /* para is one of TOUPNAM_PARA_xxx */
toupnam_ports(HRESULT)  Toupnam_get_Para(HToupNam h, unsigned para, int* value);

toupnam_ports(HRESULT)  Toupnam_put_Wifi(const char* id, ToupnamWifi wifi);
toupnam_ports(HRESULT)  Toupnam_get_Wifi(const char* id);
toupnam_ports(HRESULT)  Toupnam_list_Wifi(const char* id, void* pExtraCtx);

typedef struct {
    unsigned    type;       /* 0 => file, 1 => directory */
    char        name[256];  /* download the file with the url http://addr/path/name
                                For example, Camera's ip is 192.168.1.2, and file in the sd card directory abc/xyz.mp4, then the url is http://192.168.1.2/abc/xyz.mp4
                                So, it can be downloaded from this url with libcurl or WinInet.
                            */
}ToupnamDirItem;
toupnam_ports(HRESULT)  Toupnam_list_Dir(const char* id, const char* path, void* pExtraCtx);
toupnam_ports(HRESULT)  Toupnam_get_Thumbnail(const char* id, const char* path, void* pExtraCtx);
typedef struct {
    unsigned    type; /* 0 => add, 1 => del, 2 => rename */
    char        name[256];
    char        newname[256];
}ToupnamDirChange;
toupnam_ports(HRESULT)  Toupnam_change_Dir(const char* id, const char* path, const ToupnamDirChange* dc, unsigned dclength); /* del file or directory, rename file or directory */
toupnam_ports(HRESULT)  Toupnam_RecordStart(const char* id, const char* path, unsigned recordtime, void* pExtraCtx); /* record to the sd card */
toupnam_ports(HRESULT)  Toupnam_RecordStop(const char* id, void* pExtraCtx); /* stop record to the sd card */

toupnam_ports(HRESULT)  Toupnam_get_DateTime(const char* id, void* pExtraCtx);
toupnam_ports(HRESULT)  Toupnam_put_DateTime(const char* id, time_t t, void* pExtraCtx);

toupnam_ports(HRESULT)  Toupnam_put_Para_ById(const char* id, unsigned para, int value); /* para is one of TOUPNAM_PARA_XXX */
toupnam_ports(HRESULT)  Toupnam_get_Para_ById(const char* id, unsigned para, int* value);

typedef struct HToupnamVideoT { int unused; } *HToupnamVideo;
toupnam_ports(HToupnamVideo)    Toupnam_OpenVideo(int w, int h, int fps, int bitrate, const char* inputFile, const char* codec);
toupnam_ports(void)             Toupnam_CloseVideo(HToupnamVideo h);
toupnam_ports(HRESULT)          Toupnam_WriteVideo(HToupnamVideo h, const void* inputData, unsigned unTimeStamp);

#ifdef _WIN32
#pragma pack(pop)
#endif

#ifdef __cplusplus
}
#endif

#endif
