#ifndef __toupnam_dshow_h__
#define __toupnam_dshow_h__

/* Version: 5.13010.20180926 */

#ifdef _WIN32
#ifndef _INC_WINDOWS
#include <windows.h>
#endif
#endif

#ifndef __TOUPNAMINST_DEFINED__
#define __TOUPNAMINST_DEFINED__

#define TOUPNAM_MAX         32

typedef struct{
    int idisable;   /* 0 = "support this feature", 1 = "not support" */
    int imin; /* maximum value */
    int imax; /* minimum value */
    int idef; /* default value */
}ToupnamRange;

#define TOUPNAM_PARA_UNKNOWN                0x00
#define TOUPNAM_PARA_EXPOTIME               0x01    /* exposure time */
#define TOUPNAM_PARA_AGAIN                  0x02    /* analog gain */
#define TOUPNAM_PARA_AEXPOTARGET            0x03    /* auto exposure target */
#define TOUPNAM_PARA_TEMP                   0x04    /* color temperature */
#define TOUPNAM_PARA_TINT                   0x05
#define TOUPNAM_PARA_CONTRAST               0x06    /* contrast */
#define TOUPNAM_PARA_HUE                    0x07    /* hue */
#define TOUPNAM_PARA_SATURATION             0x08    /* saturation */
#define TOUPNAM_PARA_BRIGHTNESS             0x09    /* brightness */
#define TOUPNAM_PARA_GAMMA                  0x0a    /* gamma */
#define TOUPNAM_PARA_AEXPO                  0x0b    /* auto exposure */
#define TOUPNAM_PARA_AWB                    0x0c    /* auto white balance */
#define TOUPNAM_PARA_BINSKIP                0x0d    /* bin / skip */
#define TOUPNAM_PARA_HZ                     0x0e    /* power supply: 0 -> 60HZ AC;  1 -> 50Hz AC;   2 -> DC */
#define TOUPNAM_PARA_BPS                    0x0f    /* bits per second */
#define TOUPNAM_PARA_KEYFRAME               0x10    /* key frame interval */
#define TOUPNAM_PARA_LOWLIGHTCOMPENSATION   0x11    /* low light compensation */
#define TOUPNAM_PARA_SHARPNESS              0x12    /* sharpness */
#define TOUPNAM_PARA_WBREDGAIN              0x13    /* white balance red gain */
#define TOUPNAM_PARA_WBGREENGAIN            0x14    /* white balance green gain */
#define TOUPNAM_PARA_WBBLUEGAIN             0x15    /* white balance blue gain */
#define TOUPNAM_PARA_DENOISE                0x16    /* denoise */

#define TOUPNAM_PARA_CHROME                 0x80    /* monochromatic mode */
#define TOUPNAM_PARA_VFLIP                  0x81    /* vertical flip */
#define TOUPNAM_PARA_HFLIP                  0x82    /* horizontal flip */
#define TOUPNAM_PARA_NEGATIVE               0x83    /* negative film */

#define TOUPNAM_STATE_INITING               0x00 /* initialization */
#define TOUPNAM_STATE_NORMAL                0x01 /* normal */
#define TOUPNAM_STATE_UNREACHABLE           0x02 /* network not reachable */

#define TOUPNAM_FLAG_WIFI_AP        0x0001
#define TOUPNAM_FLAG_WIFI_STA       0x0002
#define TOUPNAM_FLAG_ETHERNET       0x0004
#define TOUPNAM_FLAG_CAPTURE        0x0008  /* support the ability of capture image from camera */
#define TOUPNAM_FLAG_AWBCHECKMODE   0x0010  /* Auto white balance: check mode vs 'one push' mode */
#define TOUPNAM_FLAG_UVC            0x0020  /* uvc camera */
#define TOUPNAM_FLAG_WBGAIN         0x0040  /* white balance gain mode or temp tint mode */
#define TOUPNAM_FLAG_MULTICAST      0x0080  /* RTSP/RTP multicast */
#define TOUPNAM_FLAG_AF             0x0100  /* support auto focus */
#define TOUPNAM_FLAG_SD_LIST        0x0200  /* support to list sd card */

typedef struct{
	char            id[64];         /* unique id, used for Toupnam_Open */
	char            sn[64];         /* serial number */
	char            name[64];
	char            model[64];
	char            version[64];
	unsigned        state;  /* TOUPNAM_STATE_xxx */
	unsigned        flag;   /* TOUPNAM_FLAG_xxx */
	ToupnamRange    range[TOUPNAM_MAX];
}ToupnamInst;

#endif

#ifndef __ITOUPNAM_DEFINED__
#define __ITOUPNAM_DEFINED__
    /* Both the DirectShow source filter and the output pin support this interface.
        That is to say, you can use QueryInterface on the filter or the pin to get the IToupnam interface. */

    // {F593CBB6-B326-4A55-9C99-B482EB1A104C}
    DEFINE_GUID(IID_IToupnam, 0xf593cbb6, 0xb326, 0x4a55, 0x9c, 0x99, 0xb4, 0x82, 0xeb, 0x1a, 0x10, 0x4c);

    DECLARE_INTERFACE_(IToupnam, IUnknown)
    {
        STDMETHOD(get_Para) (THIS_ unsigned para, int* value) PURE;
        STDMETHOD(put_Para) (THIS_ unsigned para, int value) PURE;
        STDMETHOD(get_Inst) (THIS_ ToupnamInst* p) PURE;
    };
#endif

#endif
